
# Copyright (c) Django Software Foundation and individual contributors.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# The following code was copied from the Django project, and only lightly
# modified.  Please adhere to the above copyright and license for the code
# in this file.

# Note: Nothing is covered here because this file is imported before nose and
# coverage take over.. and so its a false positive that nothing is covered.

import datetime  # pragma: nocover
import os  # pragma: nocover
import subprocess  # pragma: nocover


VERSION = (1, 0, 2, 'final', 0)


def get_version():  # pragma: nocover
    "Returns a PEP 386-compliant version number from VERSION."

    version = VERSION

    assert len(version) == 5
    assert version[3] in ('alpha', 'beta', 'rc', 'final')

    # Now build the two parts of the version number:
    # main = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|c}N - for alpha, beta and rc releases

    # We want to explicitly include all three version/release numbers
    # parts = 2 if version[2] == 0 else 3
    parts = 3
    main = '.'.join(str(x) for x in version[:parts])

    sub = ''
    if version[3] == 'alpha' and version[4] == 0:
        git_changeset = _get_git_changeset()
        if git_changeset:
            sub = '.dev%s' % git_changeset

    elif version[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
        sub = mapping[version[3]] + str(version[4])

    return main + sub


def _get_git_changeset():  # pragma: nocover
    """Returns a numeric identifier of the latest git changeset.

    The result is the UTC timestamp of the changeset in YYYYMMDDHHMMSS format.
    This value isn't guaranteed to be unique, but collisions are very
    unlikely, so it's sufficient for generating the development version
    numbers.
    """
    repo_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    git_log = subprocess.Popen('git log --pretty=format:%ct --quiet -1 HEAD',
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                               shell=True, cwd=repo_dir,
                               universal_newlines=True)
    timestamp = git_log.communicate()[0]
    try:
        timestamp = datetime.datetime.utcfromtimestamp(int(timestamp))
    except ValueError: 	# pragma: nocover
        return None  	# pragma: nocover
    return timestamp.strftime('%Y%m%d%H%M%S')



#
#   module tank.core
#

import os.path

from cement.utils import fs


def resource_path(*path_parts: str) -> str:
    """
    Resolves path to a resource.
    """
    if '..' in path_parts:
        raise ValueError('parent directory references are forbidden')

    tank_src = os.path.dirname(os.path.dirname(fs.abspath(__file__)))
    return fs.join(tank_src, 'resources', *path_parts)



#
#   module tank.core.tf
#
# Terraform-related code.
#

from os.path import dirname, isdir

from cement.utils import fs

from tank.core import resource_path
from tank.core.exc import TankError, TankTestCaseError
from tank.core.testcase import TestCase


class PlanGenerator:
    """
    Generates a Terraform plan for the run based on the testcase and the user settings.
    """

    def __init__(self, app, testcase: TestCase):
        self._app = app
        self.testcase = testcase

        if not isdir(self._provider_templates):
            raise TankError('Failed to find Terraform templates for cloud provider {} at {}'.format(
                self._app.cloud_settings.provider.value, self._provider_templates
            ))

    def generate(self, plan_dir: str):
        if self.testcase.total_instances <= 10:
            monitoring_machine_type = 'small'
        elif self.testcase.total_instances < 50:
            monitoring_machine_type = 'standard'
        else:
            monitoring_machine_type = 'large'

        self._app.template.copy(self._provider_templates, plan_dir, {
            'instances': self.testcase.instances,
            'monitoring_machine_type': monitoring_machine_type,
        })


    @property
    def _provider_templates(self) -> str:
        return resource_path('providers', self._app.cloud_settings.provider.value)



#!/usr/bin/env python

from __future__ import print_function

import os
import sys
from collections import namedtuple

from ansible.parsing.dataloader import DataLoader
from ansible.vars.manager import VariableManager
from ansible.inventory.manager import InventoryManager
from ansible.executor.playbook_executor import PlaybookExecutor

loader = DataLoader()

inventory = InventoryManager(loader=loader, sources='digital_ocean.py')
variable_manager = VariableManager(loader=loader, inventory=inventory)
playbook_path = 'play.yml'

if not os.path.exists(playbook_path):
    print('[INFO] The playbook does not exist')
    sys.exit()

Options = namedtuple('Options',
                     ['listtags', 'listtasks', 'listhosts', 'syntax',
                      'connection', 'module_path', 'forks', 'remote_user',
                      'private_key_file', 'ssh_common_args', 'ssh_extra_args',
                      'sftp_extra_args', 'scp_extra_args', 'become',
                      'become_method', 'become_user', 'verbosity', 'check',
                      'diff', 'skip_tags'])
options = Options(
    listtags=False, listtasks=False, listhosts=False, syntax=False,
    connection='ssh', module_path=None, forks=100, remote_user='root',
    private_key_file=None, ssh_common_args=None, ssh_extra_args=None,
    sftp_extra_args=None, scp_extra_args=None, become=True,
    become_method='sudo', become_user='root', verbosity=None, check=False,
    diff=False, skip_tags=['debug'])

# This can accomodate various other command line arguments.`
# variable_manager.extra_vars = {'hosts': 'mywebserver'}

passwords = {}

pbex = PlaybookExecutor(
    playbooks=[playbook_path], inventory=inventory,
    variable_manager=variable_manager, loader=loader, options=options,
    passwords=passwords)

results = pbex.run()



#
#   module tank.core.binding
#
# Binding of the Tank core to a particular blockchain.
#

import os.path
import copy
from shutil import copy2

from cement.utils import fs

from tank.core import resource_path
from tank.core.exc import TankConfigError
from tank.core.utils import yaml_load


class AnsibleBinding:
    """
    Ansible part of the binding.
    """
    BLOCKCHAIN_ROLE_NAME = 'tank.blockchain'

    def __init__(self, app, binding_name: str):
        """
        Ctor.
        :param binding_name: codename of the binding to use
        :param app: Tank app
        """
        self.binding_name = binding_name
        self._app = app

    def get_dependencies(self):
        """
        Provides ansible dependencies in the form of requirements.yml records.
        :returns: dependency record list
        """
        bindings_conf = _BindingsConfig(self._app)
        conf = bindings_conf.config.get(self.binding_name)
        if conf is None:
            raise TankConfigError('Configuration for binding named {} is not found under {}'.format(
                self.binding_name, bindings_conf.config_file
            ))

        result = {
            'src': conf['ansible']['src'],
            'name': self.__class__.BLOCKCHAIN_ROLE_NAME,
        }
        if 'version' in conf['ansible']:
            result['version'] = conf['ansible']['version']

        return [result]


class _BindingsConfig:

    def __init__(self, app):
        self._app = app

        self.config_file = fs.join(app.user_dir, 'bindings.yml')
        # TODO atomically
        if not os.path.exists(self.config_file):
            # setting up the default config
            copy2(resource_path('bindings.yml'), self.config_file)

        # TODO validate
        self._config = yaml_load(self.config_file)

    @property
    def config(self):
        return copy.deepcopy(self._config)



import logging.config
import os
from typing import Dict
import pathlib

from cement import App, TestApp, init_defaults
from cement.core.exc import CaughtSignal
from cement.utils import fs

from tank.core.cloud_settings import CloudUserSettings
from tank.core.exc import TankError
from tank.controllers.base import Base
from tank.controllers.cluster import NestedCluster, EmbeddedCluster
from tank.logging_conf import build_logging_conf


logger = logging.getLogger(__name__)


def _default_config() -> Dict:
    config = init_defaults('tank', 'log.logging')

    config['tank'] = {
        'ansible': {
            'forks': 50,
        },
    }

    config['tank']['monitoring'] = {
        "admin_user": "tank",
        "admin_password": "tank"
    }

    config['log.logging']['level'] = 'WARNING'
    return config


class MixbytesTank(App):
    """MixBytes Tank primary application."""

    class Meta:
        label = 'tank'

        # configuration defaults
        config_defaults = _default_config()

        # call sys.exit() on close
        close_on_exit = True

        # load additional framework extensions
        extensions = [
            'yaml',
            'colorlog',
            'jinja2',
        ]

        # List of configuration directory
        config_dirs = ['~/.tank']

        # configuration handler
        config_handler = 'yaml'

        # configuration file suffix
        config_file_suffix = '.yml'

        # set the log handler
        log_handler = 'colorlog'

        # set the output handler
        output_handler = 'jinja2'

        template_handler = 'jinja2'

        # register handlers
        handlers = [
            Base,
            EmbeddedCluster,
            NestedCluster,
        ]

        # register hooks
        hooks = [
        ]

    def __init__(self):
        super().__init__()
        self._cloud_settings = None

    def setup(self):
        super(MixbytesTank, self).setup()
        fs.ensure_dir_exists(self.user_dir)

        additional_config_defaults = {
            'tank': {
                'terraform_run_command': os.path.join(self.installation_dir, 'terraform'),
                'terraform_inventory_run_command': os.path.join(self.installation_dir, 'terraform-inventory'),
            },
        }

        for section, variables_dict in additional_config_defaults.items():
            for key, value in variables_dict.items():
                if key not in self.config.keys(section):
                    self.config.set(section, key, value)

    @property
    def app_env(self) -> Dict:
        env = os.environ.copy()
        env["TF_LOG"] = "TRACE"
        env["TF_IN_AUTOMATION"] = "true"
        return env

    @property
    def cloud_settings(self) -> CloudUserSettings:
        if self._cloud_settings is None:
            self._cloud_settings = CloudUserSettings(self.config)

        return self._cloud_settings

    @property
    def provider(self) -> str:
        return self.cloud_settings.provider.value

    @property
    def terraform_run_command(self) -> str:
        return self.config.get(self.Meta.label, 'terraform_run_command')

    @property
    def terraform_inventory_run_command(self) -> str:
        return self.config.get(self.Meta.label, 'terraform_inventory_run_command')

    @property
    def user_dir(self) -> str:
        return fs.abspath(fs.join(pathlib.Path.home(), '.tank'))

    @property
    def installation_dir(self) -> str:
        return fs.abspath(fs.join(self.user_dir, 'bin'))

    @property
    def ansible_config(self) -> dict:
        """Return dict with ansible parameters."""
        return self.config.get(self.Meta.label, 'ansible')


class MixbytesTankTest(TestApp, MixbytesTank):
    """A sub-class of MixbytesTank that is better suited for testing."""

    class Meta:
        label = 'tank'


def main():
    with MixbytesTank() as app:
        logs_dir = os.path.join(app.user_dir, 'logs')
        logging.config.dictConfig(build_logging_conf(logs_dir=logs_dir))

        try:
            app.run()

        except TankError as e:
            print('{}: {}'.format(e.__class__.__name__, e))
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        # FIXME better signal handling
        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)

            print(f'{e}')

            app.exit_code = 0


if __name__ == '__main__':
    main()



import logging
import os
import stat
import sys
import zipfile
from urllib.request import urlopen

import sh


_logger = logging.getLogger(__name__)


class BaseInstaller(object):
    """Base installer.

    1. Download ZIP archive.
    2. Unzip it.
    3. Remove ZIP file.
    4. Move file to storage_path directory.
    5. Make file executable
    6. Add variable to $PATH
    """

    version: str
    url: str
    archive_name: str
    file_name: str  # name of file in archive

    def __init__(self, storage_path: str):
        """Build archive full path."""
        self._storage_path = storage_path
        if not os.path.exists(storage_path):
            os.makedirs(storage_path)

        self._archive_full_path = os.path.join(storage_path, self.archive_name)
        self._file_full_path = os.path.join(storage_path, self.file_name)

    def _is_installed(self) -> bool:
        """Returns True if tool is installed else False."""
        try:
            sh.Command(self._file_full_path)
            return True
        except sh.CommandNotFound:
            return False

    def _download_archive(self):
        """Download archive from provided url."""
        _logger.debug('Downloading archive...')
        response = urlopen(self.url)

        with open(self._archive_full_path, 'wb') as archive_file:
            chunk_size = 1024 * 1024  # 1 MB
            chunk = response.read(chunk_size)

            while chunk:
                archive_file.write(chunk)
                chunk = response.read(chunk_size)

        _logger.debug('Archive {name} has been successfully downloaded.'.format(name=self.archive_name))

    def _unpack_archive(self):
        """Unpack archive with provided name."""
        with zipfile.ZipFile(self._archive_full_path, 'r') as zip_ref:
            zip_ref.extractall(self._storage_path)

        _logger.debug('Archive has been unpacked.')

    def _remove_archive(self):
        """Remove archive after unpacking."""
        os.remove(self._archive_full_path)
        _logger.debug('Archive has been removed.')

    def _make_executable(self):
        """Makes file executable."""
        file_stat = os.stat(self._file_full_path)
        os.chmod(self._file_full_path, file_stat.st_mode | stat.S_IEXEC)

    def _add_variables(self):
        """Add variables to $PATH."""
        path_variable = os.environ.get('PATH', '')
        paths = path_variable.split(os.pathsep)

        if self._storage_path not in paths:
            if path_variable:
                os.environ['PATH'] = os.pathsep.join([self._storage_path, path_variable])
            else:
                os.environ['PATH'] = self._storage_path

        _logger.debug('Variable has been added to $PATH.')

    def install(self):
        """Installation logic is here."""
        if not self._is_installed():
            _logger.debug('Installing {name}...'.format(name=self.file_name))
            self._download_archive()
            self._unpack_archive()
            self._remove_archive()
            self._make_executable()
        else:
            _logger.debug('{name} is already installed.'.format(name=self.file_name))

        self._add_variables()


class TerraformInstaller(BaseInstaller):
    """Terraform installer."""

    version = '0.11.13'
    file_name = 'terraform'
    archive_name = 'terraform_{v}_{platform}_amd64.zip'.format(v=version, platform=sys.platform.lower())
    url = 'https://releases.hashicorp.com/terraform/{v}/{filename}'.format(v=version, filename=archive_name)


class TerraformInventoryInstaller(BaseInstaller):
    """Terraform inventory installer."""

    version = 'v0.8'
    file_name = 'terraform-inventory'
    archive_name = 'terraform-inventory_{v}_{platform}_amd64.zip'.format(v=version, platform=sys.platform.lower())
    url = (
        'https://github.com/adammck/terraform-inventory/releases/download/{v}/{filename}'
    ).format(v=version, filename=archive_name)


if __name__ == '__main__':
    default_directory = os.path.join(os.path.expanduser('~'), '.tank', 'bin')
    TerraformInstaller(default_directory).install()
    TerraformInventoryInstaller(default_directory).install()



#
#   module tank.core.run
#
import os
import sys
import stat
import tempfile
from shutil import rmtree
from shutil import copytree
from time import time
from typing import Dict
from uuid import uuid4
import json
from datetime import datetime

import sh
from cement import fs
from filelock import FileLock
import namesgenerator

from tank.core import resource_path
from tank.core.binding import AnsibleBinding
from tank.core.exc import TankError, TankConfigError
from tank.core.testcase import TestCase
from tank.core.tf import PlanGenerator
from tank.core.utils import yaml_load, yaml_dump, grep_dir, json_load, sha256
from tank.terraform_installer import TerraformInstaller, TerraformInventoryInstaller


class Run:
    """
    Single run of a tank testcase.

    TODO detect and handle CloudUserSettings change.
    """

    @classmethod
    def new_run(cls, app, testcase: TestCase):
        run_id = namesgenerator.get_random_name()

        fs.ensure_dir_exists(cls._runs_dir(app))

        temp_dir = tempfile.mkdtemp(prefix='_{}'.format(run_id), dir=cls._runs_dir(app))
        cls._save_meta(temp_dir, testcase)

        # make a copy to make sure any alterations of the source won't affect us
        testcase.save(fs.join(temp_dir, 'testcase.yml'))

        copytree(resource_path('scripts'), temp_dir+'/scripts')

        # TODO prevent collisions
        os.rename(temp_dir, fs.join(cls._runs_dir(app), run_id))

        return cls(app, run_id)

    @classmethod
    def list_runs(cls, app):
        fs.ensure_dir_exists(cls._runs_dir(app))
        return [cls(app, run_id) for run_id in grep_dir(cls._runs_dir(app), '^[a-zA-Z0-9][a-zA-Z_0-9]*$', isdir=True)]


    def __init__(self, app, run_id: str):
        self._app = app
        self.run_id = run_id

        # install terraform and terraform-inventory
        TerraformInstaller(storage_path=app.installation_dir).install()
        TerraformInventoryInstaller(storage_path=app.installation_dir).install()

        self._testcase = TestCase(fs.join(self._dir, 'testcase.yml'), app)
        self._meta = yaml_load(fs.join(self._dir, 'meta.yml'))

    def init(self):
        """
        Download plugins and modules for Terraform.
        """
        with self._lock:
            self._generate_tf_plan()

            sh.Command(self._app.terraform_run_command)(
                "init", "-backend-config", "path={}".format(self._tf_state_file), self._tf_plan_dir,
                _env=self._make_env(), _out=sys.stdout, _err=sys.stderr)

    def plan(self):
        """
        Generate and show an execution plan by Terraform.
        """
        with self._lock:
            sh.Command(self._app.terraform_run_command)(
                "plan", "-input=false", self._tf_plan_dir,
                _env=self._make_env(), _out=sys.stdout, _err=sys.stderr)

    def create(self):
        """
        Create instances for the cluster.
        """
        self._check_private_key_permissions()

        with self._lock:
            sh.Command(self._app.terraform_run_command)(
                "apply", "-auto-approve", "-parallelism=51", self._tf_plan_dir,
                _env=self._make_env(), _out=sys.stdout, _err=sys.stderr)

    def dependency(self):
        """
        Install Ansible roles from Galaxy or SCM.
        """
        with self._lock:
            ansible_deps = yaml_load(resource_path('ansible', 'ansible-requirements.yml'))

            ansible_deps.extend(AnsibleBinding(self._app, self._testcase.binding).get_dependencies())

            requirements_file = fs.join(self._dir, 'ansible-requirements.yml')
            yaml_dump(requirements_file, ansible_deps)

            sh.Command("ansible-galaxy")(
                "install", "-f", "-r", requirements_file,
                _env=self._make_env(), _out=sys.stdout, _err=sys.stderr)

    def provision(self):
        self._check_private_key_permissions()

        extra_vars = {
            # including blockchain-specific part of the playbook
            'blockchain_ansible_playbook':
                fs.join(self._roles_path, AnsibleBinding.BLOCKCHAIN_ROLE_NAME, 'tank', 'playbook.yml'),
            # saving a report of the important cluster facts
            '_cluster_ansible_report': self._cluster_report_file,
            # grafana monitoring login/password
            'monitoring_user_login': self._app.cloud_settings.monitoring_vars['admin_user'],
            'monitoring_user_password': self._app.cloud_settings.monitoring_vars['admin_password'],
        }

        with self._lock:
            sh.Command("ansible-playbook")(
                "-f", self._app.ansible_config['forks'],
                "-u", "root",
                "-i", self._app.terraform_inventory_run_command,
                "--extra-vars", self._ansible_extra_vars(extra_vars),
                "--private-key={}".format(self._app.cloud_settings.provider_vars['pvt_key']),
                resource_path('ansible', 'core.yml'),
                _env=self._make_env(), _out=sys.stdout, _err=sys.stderr, _cwd=self._tf_plan_dir)

    def inspect(self):
        with self._lock:
            result = {
                'meta': self.meta,
                'testcase': self._testcase.content,
            }

            if os.path.exists(self._cluster_report_file):
                result['cluster'] = self._cluster_report()

        return result

    def bench(self, load_profile: str, tps: int, total_tx: int):
        self._check_private_key_permissions()

        bench_command = 'bench --common-config=/tool/bench.config.json ' \
                        '--module-config=/tool/blockchain.bench.config.json'
        if tps is not None:
            # It's assumed, that every node is capable of running the bench.
            per_node_tps = max(int(tps / self._testcase.total_instances), 1)
            bench_command += ' --common.tps {}'.format(per_node_tps)

        if total_tx is not None:
            # It's assumed, that every node is capable of running the bench.
            per_node_tx = max(int(total_tx / self._testcase.total_instances), 1)
            bench_command += ' --common.stopOn.processedTransactions {}'.format(per_node_tx)

        # FIXME extract hostnames from inventory, but ignore monitoring
        ips = [ip for ip, i in self._cluster_report().items() if i['bench_present']]
        if not ips:
            raise TankError('There are no nodes capable of running the bench util')
        host_patterns = ','.join(ips)

        with self._lock:
            # send the load_profile to the cluster
            extra_vars = {'load_profile_local_file': fs.abspath(load_profile)}

            sh.Command("ansible-playbook")(
                "-f", self._app.ansible_config['forks'],
                "-u", "root",
                "-i", self._app.terraform_inventory_run_command,
                "--extra-vars", self._ansible_extra_vars(extra_vars),
                "--private-key={}".format(self._app.cloud_settings.provider_vars['pvt_key']),
                "-t", "send_load_profile",
                fs.join(self._roles_path, AnsibleBinding.BLOCKCHAIN_ROLE_NAME, 'tank', 'send_load_profile.yml'),
                _env=self._make_env(), _out=sys.stdout, _err=sys.stderr, _cwd=self._tf_plan_dir)

            # run the bench
            sh.Command("ansible")(
                '-f', '150', '-B', '3600', '-P', '10', '-u', 'root',
                '-i', self._app.terraform_inventory_run_command,
                '--private-key={}'.format(self._app.cloud_settings.provider_vars['pvt_key']),
                host_patterns,
                '-a', bench_command,
                _env=self._make_env(), _out=sys.stdout, _err=sys.stderr, _cwd=self._tf_plan_dir)

    def destroy(self):
        with self._lock:
            sh.Command(self._app.terraform_run_command)(
                "destroy", "-auto-approve", "-parallelism=100",
                self._tf_plan_dir,
                _env=self._make_env(), _out=sys.stdout, _err=sys.stderr)

            # atomic move before cleanup
            temp_dir = fs.join(self.__class__._runs_dir(self._app), '_{}'.format(self.run_id))
            os.rename(self._dir, temp_dir)

        # cleanup with the lock released
        rmtree(temp_dir)


    @property
    def meta(self) -> Dict:
        return dict(self._meta)

    @property
    def created_at(self) -> datetime:
        return datetime.fromtimestamp(self.meta['created'])

    @property
    def testcase_copy(self) -> TestCase:
        """
        Copy of the original testcase.
        """
        return self._testcase


    @classmethod
    def _runs_dir(cls, app) -> str:
        return fs.join(app.user_dir, 'run')

    @classmethod
    def _save_meta(cls, run_dir: str, testcase: TestCase):
        yaml_dump(fs.join(run_dir, 'meta.yml'), {
            'testcase_filename': fs.abspath(testcase.filename),
            'created': int(time()),
            'setup_id': sha256(uuid4().bytes)[:12],
        })

    def _ansible_extra_vars(self, extra: Dict = None) -> str:
        a_vars = dict(('bc_{}'.format(k), str(v)) for k, v in self._app.cloud_settings.ansible_vars.items())
        a_vars.update(dict(('bc_{}'.format(k), str(v)) for k, v in self._testcase.ansible.items()))

        if extra is not None:
            a_vars.update(extra)

        return json.dumps(a_vars, sort_keys=True)

    def _make_env(self) -> Dict:
        fs.ensure_dir_exists(self._tf_data_dir)
        fs.ensure_dir_exists(self._log_dir)

        env = self._app.app_env

        env["TF_LOG_PATH"] = fs.join(self._log_dir, 'terraform.log')
        env["TF_DATA_DIR"] = self._tf_data_dir
        env["TF_VAR_state_path"] = self._tf_state_file
        env["TF_VAR_blockchain_name"] = self._testcase.binding.replace('_', '-')[:10]
        env["TF_VAR_setup_id"] = self._meta['setup_id']
        env["TF_VAR_scripts_path"] = fs.join(self._dir, 'scripts')

        for k, v in self._app.cloud_settings.provider_vars.items():
            env["TF_VAR_{}".format(k)] = v

        env["ANSIBLE_ROLES_PATH"] = self._roles_path
        env["ANSIBLE_CONFIG"] = resource_path('ansible', 'ansible.cfg')
        env["ANSIBLE_LOG_PATH"] = fs.join(self._log_dir, 'ansible.log')

        return env

    def _generate_tf_plan(self):
        """
        Generation of Terraform manifests specific for this run and user preferences.
        """
        PlanGenerator(self._app, self._testcase).generate(self._tf_plan_dir)

    def _cluster_report(self):
        return json_load(self._cluster_report_file)

    def _check_private_key_permissions(self):
        """
        Checks whether groups and others have 0 access to private key
        """
        # oct -'0o77', bin - '0b000111111', which is the same as ----rwxrwx
        NOT_OWNER_PERMISSION = stat.S_IRWXG + stat.S_IRWXO

        file_stat: os.stat_result = os.stat(self._app.cloud_settings.provider_vars['pvt_key'])
        file_mode = stat.S_IMODE(file_stat.st_mode)

        if file_mode & NOT_OWNER_PERMISSION != 0:
            raise TankConfigError('Private key has wrong permission mask.')

    @property
    def _dir(self) -> str:
        return fs.join(self.__class__._runs_dir(self._app), self.run_id)

    @property
    def _lock(self) -> FileLock:
        return FileLock(fs.join(self._dir, '.lock'))

    @property
    def _tf_data_dir(self) -> str:
        return fs.join(self._dir, 'tf_data')

    @property
    def _tf_plan_dir(self) -> str:
        return fs.join(self._dir, 'tf_plan')

    @property
    def _tf_state_file(self) -> str:
        return fs.join(self._dir, "blockchain.tfstate")

    @property
    def _log_dir(self) -> str:
        return fs.join(self._dir, 'log')

    @property
    def _roles_path(self) -> str:
        return fs.join(self._dir, "ansible_roles")

    @property
    def _cluster_report_file(self) -> str:
        return fs.join(self._dir, 'cluster_ansible_report.json')



#
#   module tank.core.lambda
#


def first(subscriptable):
    return subscriptable[0]


def second(subscriptable):
    return subscriptable[1]


def third(subscriptable):
    return subscriptable[2]



#
#   module tank.core.testcase
#
import copy
from typing import List

from jsonschema import Draft4Validator, ValidationError

from tank.core import resource_path
from tank.core.exc import TankTestCaseError
from tank.core.regions import RegionsConfig
from tank.core.utils import yaml_load, yaml_dump, ratio_from_percent, split_evenly


class InstancesCanonizer(object):
    """Canonize config for instances.

    Accept only valid instances content.
    Transforms dict to following format:

    Role:
      Region1:
        count: ...
        type: ...
        packetloss: ...
      Region2:
        count: ...
        type: ...
        packetloss: ...
    """

    _GENERAL_OPTIONS = {
        'type': 'small',
        'packetloss': 0,
    }

    def __init__(self, instances_content: dict):
        """Load content and defaults."""
        self._content = instances_content
        self._global_defaults = self._load_defaults()

    def _load_defaults(self) -> dict:
        """Provide defaults for config.

        Now they are default region and type.
        """
        defaults = dict()

        for option, default in self._GENERAL_OPTIONS.items():
            defaults[option] = self._content.get(option, default)
            if option in self._content.keys():
                self._content.pop(option)

        defaults['region'] = 'default'
        return defaults

    def _build_configuration(self, count: int, type: str = None, packetloss: int = None) -> dict:
        """Build minimal configuration from parameters."""
        configuration = {
            'count': count,
            'type': self._global_defaults['type'] if type is None else type,
            'packetloss': ratio_from_percent(self._global_defaults['packetloss'] if packetloss is None else packetloss),
        }

        return configuration

    def canonize(self) -> dict:
        """Convert to canonized config."""
        canonized_dict = dict()

        for role, config in self._content.items():
            if isinstance(config, int):  # shortest number configuration
                canonized_dict[role] = {
                    self._global_defaults['region']: self._build_configuration(count=config),
                }
            elif 'regions' in config:  # dict configuration with regions
                canonized_dict[role] = dict()

                for region, region_config in config['regions'].items():
                    if isinstance(region_config, int):
                        canonized_dict[role][region] = self._build_configuration(
                            count=region_config,
                            type=config.get('type'),
                            packetloss=config.get('packetloss'),
                        )
                    else:
                        canonized_dict[role][region] = self._build_configuration(
                            count=region_config['count'],
                            type=region_config.get('type', config.get('type')),
                            packetloss=region_config.get('packetloss', config.get('packetloss')),
                        )
            else:  # dict configuration without regions (config must contain count param in this case)
                canonized_dict[role] = {
                    self._global_defaults['region']: self._build_configuration(
                        count=config['count'],
                        type=config.get('type'),
                        packetloss=config.get('packetloss'),
                    )
                }

        return canonized_dict


class RegionsConverter(object):
    """Convert and merge regions in canonized instances configuration.

    Convert regions to machine-readable format.
    random = equal amount for all regions.

    Example below:
        Role:
            default:
                count: 1
                type: small
                packetloss: 0
        ...

    Convert to:
        Role:
            - {
                region: FRA1
                count: 1
                type: small
                packetloss: 0
            }
            - ...
            - ...
        ...
    """

    _GROUP_PARAMETERS = ('region', 'type', 'packetloss',)

    def __init__(self, app):
        """Save provider, load RegionsConfig."""
        self._provider = app.provider
        self._regions_config = RegionsConfig(app).config
        self._available_regions = RegionsConfig.REGIONS

    def _merge_configurations(self, machine_configurations: List[dict]) -> List[dict]:
        """Merge machine configurations with equal parameters."""
        configurations_dict = dict()

        for configuration in machine_configurations:
            key = tuple(configuration[param] for param in self._GROUP_PARAMETERS)

            if key in configurations_dict:
                configurations_dict[key]['count'] += configuration['count']
            else:
                configurations_dict[key] = configuration

        return list(configurations_dict.values())

    def _convert_region(self, human_readable: str) -> str:
        """Convert region from human readable type to machine readable via regions config."""
        return self._regions_config[self._provider][human_readable]

    def convert(self, instances_config: dict) -> dict:
        """Convert configuration to machine readable."""
        converted_config = dict()

        for role, config in instances_config.items():
            machines_configurations = []

            for region, region_config in config.items():
                if region == 'random':
                    for i, count in enumerate(split_evenly(region_config['count'], len(self._available_regions))):
                        if count:
                            machines_configurations.append(
                                {
                                    'region': self._convert_region(self._available_regions[i]),
                                    'count': count,
                                    'packetloss': region_config['packetloss'],
                                    'type': region_config['type'],
                                }
                            )
                else:
                    machines_configurations.append(
                        {
                            'region': self._convert_region(region),
                            **region_config
                        },
                    )

            converted_config[role] = self._merge_configurations(machines_configurations)

        return converted_config


class TestCaseValidator(object):
    """Class for validation TestCase object."""

    SCHEMA_FILE = resource_path('testcase_schema.yml')

    def __init__(self, content: dict, filename):
        """Save filename with content."""
        self._content = content
        self._filename = filename

    def validate(self):
        """Perform all validations."""
        self._validate_schema()
        self._check_reserved_names()
        self._check_counts_equality()

    def _validate_schema(self):
        """Validate via JSON schema."""
        try:
            Draft4Validator(yaml_load(self.SCHEMA_FILE)).validate(self._content)
        except ValidationError as e:
            raise TankTestCaseError('Failed to validate testcase {}'.format(self._filename), e)

    def _check_reserved_names(self):
        """Check reserved names existence."""
        reserved_names = {'count', 'monitoring'}

        for role in self._content['instances'].keys():
            if role.lower() in reserved_names:
                raise TankTestCaseError('\'{name}\' instance name is reserved'.format(name=role))

    def _check_counts_equality(self):
        """Check if total count is equal with sum of counts in regions."""
        for role, config in self._content['instances'].items():
            if isinstance(config, dict) and all(key in config for key in ['count', 'regions']):
                total_count = config['count']
                regions_count = 0

                for region, region_config in config['regions'].items():
                    if isinstance(region_config, int):
                        regions_count += region_config
                    else:
                        regions_count += region_config['count']

                if total_count != regions_count:
                    raise TankTestCaseError(
                        'The total count does not match sum of count in regions in role {}'.format(role)
                    )


class TestCase(object):
    """Entity describing single test performed by Tank."""

    def __init__(self, filename, app):
        """Load content."""
        self._app = app
        self._filename = filename
        self._original_content = yaml_load(filename)

        self._content = copy.deepcopy(self._original_content)
        TestCaseValidator(self._content, filename).validate()
        self._content = self._prepare_content()

    @property
    def filename(self) -> str:
        return self._filename

    @property
    def binding(self) -> str:
        """Return provided binding."""
        return self._content['binding']

    @property
    def instances(self) -> dict:
        """Return copy of instances."""
        return copy.deepcopy(self._content['instances'])

    @property
    def total_instances(self) -> int:
        """Calculate amount of all instances.

        It works only after instances config canonization and converting.
        """
        instances_amount = 0
        for config in self._content['instances'].values():
            for item in config:
                instances_amount += item['count']

        return instances_amount

    @property
    def ansible(self) -> dict:
        """Return copy of ansible config."""
        return copy.deepcopy(self._content['ansible'])

    @property
    def content(self) -> dict:
        """Return copy of all content."""
        return copy.deepcopy(self._content)

    def save(self, filename):
        """Save original content to file."""
        yaml_dump(filename, self._original_content)

    def _prepare_content(self):
        """Convert to canonized config."""
        result = dict()
        canonized_instances = InstancesCanonizer(self._content['instances']).canonize()
        result['instances'] = RegionsConverter(self._app).convert(canonized_instances)
        result['binding'] = self._content['binding']
        result['ansible'] = self._content.get('ansible', dict())
        return result



import os


def build_logging_conf(logs_dir: str) -> dict:
    """Builds dict logging config."""
    if not os.path.exists(logs_dir):
        os.makedirs(logs_dir)

    logging_conf = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'simple': {
                'format': '%(asctime)s %(name)s(%(lineno)d) - %(levelname)s: %(message)s',
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': 'WARNING',
                'formatter': 'simple',
                'stream': 'ext://sys.stdout',
            },
            'info_file_handler': {
                'class': 'logging.handlers.RotatingFileHandler',
                'level': 'DEBUG',
                'formatter': 'simple',
                'filename': os.path.join(logs_dir, 'info.log'),
                'maxBytes': 10 * 1024 * 1024,  # 10 MB
                'backupCount': 3,
                'encoding': 'utf8',
            },
            'error_file_handler': {
                'class': 'logging.handlers.RotatingFileHandler',
                'level': 'ERROR',
                'formatter': 'simple',
                'filename': os.path.join(logs_dir, 'error.log'),
                'maxBytes': 10 * 1024 * 1024,  # 10 MB
                'backupCount': 3,
                'encoding': 'utf8',
            },
        },
        'root': {
            'level': 'DEBUG',
            'handlers': ['console', 'info_file_handler', 'error_file_handler'],
        },
    }

    return logging_conf



#
#   module tank.core.cloud_settings
#

from enum import Enum

import yaml
import jsonschema

from tank.core.exc import TankConfigError


class CloudProvider(Enum):
    DIGITAL_OCEAN = 'digitalocean'
    GOOGLE_CLOUD_ENGINE = 'gce'

    def __repr__(self):
        return '<%s.%s>' % (self.__class__.__name__, self.name)

    def __str__(self):
        """
        Serializes a member into string.
        """
        return self.value

    @classmethod
    def from_string(cls, str_value):
        """
        Deserializes a member from string.
        :param str_value: serialized form
        :return: enum member or None if not found
        """
        for m in cls:
            if m.value == str_value:
                return m
        else:
            return None


class CloudUserSettings:
    """
    Management and validation of cloud provider user-specific settings.
    """

    def __init__(self, app_config):
        self.provider = CloudProvider.from_string(app_config.get('tank', 'provider'))
        if self.provider is None:
            raise TankConfigError('Cloud provider is not specified or not known')

        self.monitoring_vars = app_config.get_dict()['tank'].get('monitoring')
        try:
            jsonschema.validate(self.monitoring_vars, self.__class__._MONITORING_SCHEMA)
        except jsonschema.ValidationError as e:
            raise TankConfigError('Failed to validate admin_user/password monitoring settings', e)

        self.provider_vars = app_config.get_dict().get(self.provider.value)
        if self.provider_vars is None or not isinstance(self.provider_vars, dict):
            raise TankConfigError('Cloud provider is not configured')

        self.ansible_vars = self.provider_vars.pop('ansible', dict())

        try:
            jsonschema.validate(self.provider_vars, self.__class__._SCHEMAS[self.provider])
        except jsonschema.ValidationError as e:
            raise TankConfigError('Failed to validate config for cloud provider {}'.format(self.provider), e)

        assert 'pvt_key' in self.provider_vars, 'pvt_key is required for ansible'

        try:
            jsonschema.validate(self.ansible_vars, self.__class__._ANSIBLE_SCHEMA)
        except jsonschema.ValidationError as e:
            raise TankConfigError('Failed to validate ansible config for cloud provider {}'.format(self.provider), e)

        if 'private_interface' not in self.ansible_vars:
            self.ansible_vars['private_interface'] = {
                CloudProvider.DIGITAL_OCEAN: 'eth0',
                CloudProvider.GOOGLE_CLOUD_ENGINE: 'ens4',
            }[self.provider]


    _SCHEMAS = {
        CloudProvider.DIGITAL_OCEAN: yaml.safe_load(r'''
type: object
additionalProperties: False
required:
    - token
    - pvt_key
    - ssh_fingerprint
properties:
    token:
        type: string
    pvt_key:
        type: string
    ssh_fingerprint:
        type: string
'''),

        CloudProvider.GOOGLE_CLOUD_ENGINE: yaml.safe_load(r'''
type: object
additionalProperties: False
required:
    - pub_key
    - pvt_key
    - cred_path
    - project
properties:
    pub_key:
        type: string
    pvt_key:
        type: string
    cred_path:
        type: string
    project:
        type: string
'''),
    }

    _ANSIBLE_SCHEMA = yaml.safe_load(r'''
type: object
additionalProperties: False
properties:
    private_interface:
        type: string
''')

    _MONITORING_SCHEMA = yaml.safe_load(r'''
type: object
additionalProperties: False
required:
    - admin_user
    - admin_password
properties:
    admin_user:
        type: string
    admin_password:
        type: string
''')




#
#   module tank.core.utils
#
# Misc. utils.
#
import os
import re
import json
import hashlib
from typing import List

import yaml


def yaml_load(filename: str):
    with open(filename) as fh:
        return yaml.safe_load(fh)


def yaml_dump(filename: str, data):
    with open(filename, 'w') as fh:
        return yaml.dump(data, fh, default_flow_style=False)


def json_load(filename: str):
    with open(filename) as fh:
        return json.load(fh)


def sha256(bin_data) -> str:
    return hashlib.sha256(bin_data).hexdigest()


def grep_dir(dirname: str, filter_regex: str = None, isdir: bool = False):
    """
    Enumerate and filter contents of a directory.
    """
    contents = os.listdir(dirname)

    if filter_regex is not None:
        filter_re = re.compile(filter_regex)
        contents = filter(lambda name: filter_re.match(name) is not None, contents)

    if isdir is not None:
        contents = filter(lambda name: os.path.isdir(os.path.join(dirname, name)), contents)

    return contents


def ratio_from_percent(percent: int) -> float:
    """Convert percent to ratio."""
    return percent / 100


def split_evenly(number: int, count: int) -> List[int]:
    """Return mostly equal parts.

    Example: number = 11, count = 3, result = [4, 4, 3]
    """
    parts = []

    for _ in range(count):
        if number % count:
            parts.append(number // count + 1)
            number -= parts[-1]
        else:
            parts.append(number // count)
            number -= parts[-1]

        count -= 1

    return parts



import copy
import os
from shutil import copy2

from cement.utils import fs

from tank.core import resource_path
from tank.core.utils import yaml_load


class RegionsConfig(object):
    """Config object for regions."""

    FILE_NAME = 'regions.yml'
    REGIONS = ('Europe', 'Asia', 'NorthAmerica',)

    def __init__(self, app):
        """Load or copy config file."""
        self._config_file = fs.join(app.user_dir, self.FILE_NAME)

        if not os.path.exists(self._config_file):
            copy2(resource_path('regions.yml'), self._config_file)

        self._config = yaml_load(self._config_file)

    @property
    def config(self):
        """Return loaded config."""
        return copy.deepcopy(self._config)




from setuptools import setup, find_packages
from tank.version import get_version

with open('README.md', 'r') as f:
    LONG_DESCRIPTION = f.read()

with open('requirements.txt', 'r') as f:
    requires = list(filter(None, map(str.strip, f.read().split('\n'))))

setup_options = dict(
    name='mixbytes-tank',
    version=get_version(),
    description='Benchmark engine for blockchains',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='MixBytes LLC',
    author_email='hello@mixbytes.io',
    url='https://github.com/mixbytes/tank/',
    license='Apache-2.0',
    classifiers=[
        # https://pypi.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Topic :: Software Development :: Testing",
        "Topic :: System :: Benchmark",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Clustering",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython"
    ],
    keywords='bench benchmark blockchain',
    packages=find_packages(exclude=['ez_setup', 'tests*']),
    include_package_data=True,
    install_requires=requires,
    python_requires='>=3',
    entry_points="""
        [console_scripts]
        tank = tank.main:main
    """,
)

setup(**setup_options)




class TankError(RuntimeError):
    """
    Base class of Tank-specific errors.
    """
    pass


class TankConfigError(TankError):
    """
    Configuration error.
    """
    pass


class TankTestCaseError(TankError):
    """
    Error specific for a test case configuration.
    """
    pass




import sys
import json

from cement import Controller, ex
from tabulate import tabulate

from tank.core.exc import TankError
from tank.core.run import Run
from tank.core.testcase import TestCase
from tank.core.lambdas import first, second


class BaseClusterController(Controller):
    """Base controller for overriding."""

    def _deploy(self):
        testcase = TestCase(first(self.app.pargs.testcase), self.app)
        run = Run.new_run(self.app, testcase)
        print('Created tank run: {}'.format(run.run_id))

        run.init()
        run.create()
        run.dependency()
        run.provision()

        self._show_hosts(run.inspect())
        print('\nTank run id: {}'.format(run.run_id))

    def _show_hosts(self, run_inspect_data):
        if 'cluster' not in run_inspect_data:
            raise TankError('There are no information about hosts. Have you performed provision/deploy?')

        rows = sorted([[ip, i['hostname']] for ip, i in run_inspect_data['cluster'].items()], key=second)
        print(tabulate(list(rows), headers=['IP', 'HOSTNAME']))

        for ip, info in run_inspect_data['cluster'].items():
            if info['hostname'].endswith('-monitoring'):
                print('\nMonitoring: http://{}/'.format(ip))
                break


class EmbeddedCluster(BaseClusterController):
    """Embedded cluster controller for providing short commands."""

    class Meta:
        label = 'embedded cluster'
        stacked_type = 'embedded'
        stacked_on = 'base'

    @ex(
        help='Create and setup a cluster (init, create, dependency, provision)',
        arguments=[(['testcase'], {'type': str, 'nargs': 1})]
    )
    def run(self):
        """Equal with cluster deploy."""
        self._deploy()


class NestedCluster(BaseClusterController):

    class Meta:
        label = 'cluster'
        stacked_type = 'nested'
        stacked_on = 'base'

        # text displayed at the top of --help output
        description = 'Manipulating a cluster'

        # text displayed at the bottom of --help output
        title = 'Low level cluster management commands'
        help = 'Low level cluster management commands'

    @ex(help='Show clusters')
    def list(self):
        runs = Run.list_runs(self.app)

        def make_row(run):
            return [
                run.run_id,
                run.created_at.strftime('%c'),
                run.testcase_copy.total_instances + 1,
                run.meta['testcase_filename']
            ]

        print(tabulate(list(map(make_row, runs)), headers=['RUN ID', 'CREATED', 'INSTANCES', 'TESTCASE']))

    @ex(help='Init a Tank run, download plugins and modules for Terraform', hide=True,
        arguments=[(['testcase'], {'type': str, 'nargs': 1})])
    def init(self):
        testcase = TestCase(first(self.app.pargs.testcase), self.app)
        run = Run.new_run(self.app, testcase)
        print('Created tank run: {}'.format(run.run_id))

        run.init()

    @ex(help='Generate and show an execution plan by Terraform', hide=True,
        arguments=[(['run_id'], {'type': str, 'nargs': 1})])
    def plan(self):
        Run(self.app, first(self.app.pargs.run_id)).plan()

    @ex(help='Create instances for cluster', hide=True,
        arguments=[(['run_id'], {'type': str, 'nargs': 1})])
    def create(self):
        Run(self.app, first(self.app.pargs.run_id)).create()

    @ex(help='Install Ansible roles from Galaxy or SCM', hide=True,
        arguments=[(['run_id'], {'type': str, 'nargs': 1})])
    def dependency(self):
        Run(self.app, first(self.app.pargs.run_id)).dependency()

    @ex(help='Setup instances: configs, packages, services, etc', hide=True,
        arguments=[(['run_id'], {'type': str, 'nargs': 1})])
    def provision(self):
        Run(self.app, first(self.app.pargs.run_id)).provision()

    @ex(help='Runs bench on prepared cluster',
        arguments=[
            (['run_id'],
             {'type': str, 'nargs': 1}),
            (['load_profile'],
             {'type': str, 'nargs': 1}),
            (['--tps'],
             {'help': 'set global transactions per second generation rate',
              'type': int}),
            (['--total-tx'],
             {'help': 'how many transactions to send',
              'type': int}),
        ])
    def bench(self):
        Run(self.app, first(self.app.pargs.run_id)).bench(
            first(self.app.pargs.load_profile), self.app.pargs.tps, self.app.pargs.total_tx)

    @ex(help='Destroy all instances of the cluster',
        arguments=[(['run_id'], {'type': str, 'nargs': 1})])
    def destroy(self):
        Run(self.app, first(self.app.pargs.run_id)).destroy()

    @ex(help='Low-level info about a run',
        arguments=[(['run_id'], {'type': str, 'nargs': 1})])
    def inspect(self):
        data = Run(self.app, first(self.app.pargs.run_id)).inspect()
        json.dump(data, sys.stdout, indent=4, sort_keys=True)

    @ex(help='Info about a run',
        arguments=[(['info_type'], {'choices': ['hosts'], 'nargs': 1}),
                   (['run_id'], {'type': str, 'nargs': 1})])
    def info(self):
        info_type = first(self.app.pargs.info_type)
        data = Run(self.app, first(self.app.pargs.run_id)).inspect()

        if info_type == 'hosts':
            self._show_hosts(data)

    @ex(
        help='Create and setup a cluster (init, create, dependency, provision)',
        arguments=[(['testcase'], {'type': str, 'nargs': 1})]
    )
    def deploy(self):
        self._deploy()




from cement import Controller
from cement.utils.version import get_version_banner

from tank.version import get_version

VERSION_BANNER = """
Bench toolkit for blockchain %s
%s
""" % (get_version(), get_version_banner())


class Base(Controller):
    class Meta:
        label = 'base'

        # text displayed at the top of --help output
        description = 'Bench toolkit for blockchain'

        # controller level arguments. ex: 'tank --version'
        arguments = [
            # add a version banner
            (
                ['-v', '--version'],
                {'action': 'version', 'version': VERSION_BANNER},
            ),
        ]

    def _default(self):
        """Default action if no sub-command is passed."""
        self.app.args.print_help()
